############################################################################

# Created by: Prof. Valdecy Pereira, D.Sc.
# UFF - Universidade Federal Fluminense (Brazil)
# email:  valdecy.pereira@gmail.com
# Lesson: pyCombinatorial - Genetic Algorithm
  
# GitHub Repository: <https://github.com/Valdecy>

############################################################################

# Required Libraries
import copy
import numpy  as np
import random
import os

############################################################################

# Function
def target_function():
    return

############################################################################

# Function: 2_opt
def local_search_2_opt(city_tour, recursive_seeding, target_function = target_function):
    if (recursive_seeding < 0):
        count = -2
    else:
        count = 0
    city_list = copy.deepcopy(city_tour)
    distance  = city_list[1]*2
    while (count < recursive_seeding):
        best_route = copy.deepcopy(city_list)
        seed       = copy.deepcopy(city_list)        
        for i in range(0, len(city_list[0]) - 1):
            for j in range(i+1, len(city_list[0])):
                best_route[0][i:j+1] = list(reversed(best_route[0][i:j+1]))               
                best_route[1]        = target_function(best_route[0])                    
                if (city_list[1] > best_route[1]):
                    city_list = copy.deepcopy(best_route)         
                best_route = copy.deepcopy(seed)
        count = count + 1
        if (distance > city_list[1] and recursive_seeding < 0):
             distance          = city_list[1]
             count             = -2
             recursive_seeding = -1
        elif(city_list[1] >= distance and recursive_seeding < 0):
            count             = -1
            recursive_seeding = -2
    return city_list

############################################################################

# Function: Initial Seed
def seed_function(jobs, target_function = target_function):
    seed     = [[],float("inf")]
    sequence = random.sample(list(range(0, jobs)), jobs)
    seed[0]  = sequence
    seed[1]  = target_function(seed[0])
    return seed

# Function: Initial Population
def initial_population(population_size, jobs, target_function = target_function):
    population = []
    for i in range(0, population_size):
        seed = seed_function(jobs, target_function)
        population.append(seed)
    return population

############################################################################

# Function: Fitness
def fitness_function(cost, population_size): 
    fitness = np.zeros((population_size, 2))
    for i in range(0, fitness.shape[0]):
        fitness[i,0] = 1/(1 + cost[i] + abs(np.min(cost)))
    fit_sum      = fitness[:,0].sum()
    fitness[0,1] = fitness[0,0]
    for i in range(1, fitness.shape[0]):
        fitness[i,1] = (fitness[i,0] + fitness[i-1,1])
    for i in range(0, fitness.shape[0]):
        fitness[i,1] = fitness[i,1]/fit_sum
    return fitness

# Function: Selection
def roulette_wheel(fitness): 
    ix     = 0
    random = int.from_bytes(os.urandom(8), byteorder = 'big') / ((1 << 64) - 1)
    for i in range(0, fitness.shape[0]):
        if (random <= fitness[i, 1]):
          ix = i
          break
    return ix

# Function: TSP Crossover - BCR (Best Cost Route Crossover)
def crossover_tsp_bcr(parent_1, parent_2, target_function = target_function):
    individual = copy.deepcopy(parent_2)
    cut        = random.sample(list(range(0, len(parent_1[0]))), int(len(parent_1)/2))
    cut        = [ parent_1[0][i] for i in cut ]
    d_1        = float('+inf')
    for j in range(0, len(cut)):
        best      = []
        A         = cut[j]
        parent_2[0].remove(A)
        dist_list = []
        for i in range(0, len(parent_2) + 1):
            new_list = parent_2[0][:i] + [A] + parent_2[0][i:]
            dist_list.append(new_list)
        fun_list  = [target_function( item ) for item in dist_list]
        d_2       = min(fun_list)
        if (d_2 <= d_1):
            d_1        = d_2
            n          = fun_list.index(d_1)
            best       = dist_list[n]
            individual = [best, d_1]
    return individual

# Function: Breeding
def breeding(population, fitness, elite, target_function = target_function):
    cost = [item[1] for item in population]
    if (elite > 0):
        cost, offspring = (list(t) for t in zip(*sorted(zip(cost, population))))
    for i in range (elite, len(offspring)):
        parent_1, parent_2 = roulette_wheel(fitness), roulette_wheel(fitness)
        while parent_1 == parent_2:
            parent_2 = random.sample(range(0, len(population) - 1), 1)[0]
        parent_1 = copy.deepcopy(population[parent_1])  
        parent_2 = copy.deepcopy(population[parent_2])
        rand     = int.from_bytes(os.urandom(8), byteorder = 'big') / ((1 << 64) - 1)  
        if (rand > 0.5):
            offspring[i] = crossover_tsp_bcr(parent_1, parent_2, target_function)
        else:
            offspring[i] = crossover_tsp_bcr(parent_2, parent_1, target_function)
    return offspring

# Function: Mutation - Swap with 2-opt Local Search
def mutation_tsp_swap(individual, target_function = target_function):
    k                 = random.sample(list(range(0, len(individual[0]))), 2)
    k1                = k[0]
    k2                = k[1]  
    A                 = individual[0][k1]
    B                 = individual[0][k2]
    individual[0][k1] = B
    individual[0][k2] = A
    individual[1]     = target_function(individual[0])
    individual        = local_search_2_opt(individual, -1, target_function)
    return individual

# Function: Mutation
def mutation(offspring, mutation_rate, elite, target_function):
    for i in range(elite, len(offspring)):
        probability = int.from_bytes(os.urandom(8), byteorder = 'big') / ((1 << 64) - 1)
        if (probability <= mutation_rate):
            offspring[i] = mutation_tsp_swap(offspring[i], target_function)
    return offspring

############################################################################

# Function: GA
def genetic_algorithm(jobs, population_size, elite, mutation_rate, generations, target_function = target_function, verbose = True):
    population       = initial_population(population_size, jobs, target_function)
    cost             = [item[1] for item in population]
    cost, population = (list(t) for t in zip(*sorted(zip(cost, population))))
    elite_ind        = population[0] 
    fitness          = fitness_function(cost, population_size)
    count            = 0
    while (count <= generations): 
        if (verbose == True):
            print('Generation: ', count)
        offspring        = breeding(population, fitness, elite, target_function)  
        offspring        = mutation(offspring, mutation_rate, elite, target_function)
        cost             = [item[1] for item in offspring]
        cost, population = (list(t) for t in zip(*sorted(zip(cost, offspring ))))
        elite_child      = population[0]
        fitness          = fitness_function(cost, population_size)
        if(elite_ind[1] > elite_child[1]):
            elite_ind = elite_child 
        count = count + 1  
    route, distance = elite_ind
    return route, distance

############################################################################


############################################################################

# Created by: Prof. Valdecy Pereira, D.Sc.
# UFF - Universidade Federal Fluminense (Brazil)
# email:  valdecy.pereira@gmail.com
# Elitist Combinatorial Multiobjective Optimization Algorithm

# Citation: 
# PEREIRA, V. (2024). Project: GA Scheduler, GitHub repository: <https://github.com/Valdecy/GA_Scheduler>

############################################################################

# Required Libraries
import copy
import numpy  as np
import random
import os

############################################################################

# Function 1
def func_1():
    return

# Function 2
def func_2():
    return

############################################################################

# Function: Initial Seed
def seed_function_ecmoa(jobs, list_of_functions):
    seed     = [[]] + [float("inf") for item in list_of_functions]
    sequence = random.sample((range(0, jobs)), jobs)
    seed[0]  = sequence
    for k in range (1, len(list_of_functions) + 1):
        seed[-k] = list_of_functions[-k](seed[0])
    return seed

# Function: Initial Population
def initial_population_ecmoa(population_size, jobs, list_of_functions):
    population = []
    for i in range(0, population_size):
        seed = seed_function_ecmoa(jobs, list_of_functions)
        population.append(seed)
    return population

############################################################################

# Function: Crowding Distance (Adapted from PYMOO)
def crowding_distance_function(population, M):
    position = [ (item[1:]) for item in population]
    position = np.asarray(position)/1.0
    if (position.shape[0] <= 2):
        return np.full( position.shape[0], float('+inf'))
    else:
        arg_1    = np.argsort( position, axis = 0, kind = 'mergesort')
        position = position[arg_1, np.arange(M)]
        dist     = np.concatenate([ position, np.full((1, M), np.inf)]) - np.concatenate([np.full((1, M), -np.inf), position])
        idx      = np.where(dist == 0)
        a        = np.copy(dist)
        b        = np.copy(dist)
        for i, j in zip(*idx):
            a[i, j] = a[i - 1, j]
        for i, j in reversed(list(zip(*idx))):
            b[i, j] = b[i + 1, j]
        norm            = np.max( position, axis = 0) - np.min(position, axis = 0)
        norm[norm == 0] = np.nan
        a, b            = a[:-1]/norm, b[1:]/norm
        a[np.isnan(a)]  = 0.0
        b[np.isnan(b)]  = 0.0
        arg_2           = np.argsort(arg_1, axis = 0)
        crowding        = np.sum(a[arg_2, np.arange(M)] + b[arg_2, np.arange(M)], axis = 1) / M
    crowding[np.isinf(crowding)] = float('+inf')
    crowding                     = crowding.reshape((-1,1))
    return crowding

############################################################################

# Function: Unique individuals
def ensure_unique_population(population):
    unique_population = set()
    for item in population:
        unique_population.add(tuple(item[0])) 
    unique_population_list = []
    for unique_item in unique_population:
        for original_item in population:
            if (list(unique_item) == original_item[0]):
                unique_population_list.append(original_item)
                break  
    return unique_population_list

# Function: Leaders Selection
def selection_leaders(size, M, population):   
    position = [[] for item in population]
    for m in range(0, len(population)):
        for n in range(1, len(population[m])):
            position[m].append((population[m][n]))
    position = np.array(position)
    idx      = pareto_front_points(position, pf_min = True)
    if (len(idx) > 0):
        leaders = [population[i] for i in range(0, len(idx)) if idx[i] == True]
    leaders = ensure_unique_population(leaders)
    if (size == 1):
        leaders = random.choice(leaders)
        leaders = [leaders]
    if (len(leaders) > size and size > 1):
        leaders = [leaders[i] for i in range(0, size)]  
    return leaders

# Function: TSP Crossover - BCR (Best Cost Route Crossover)
def crossover_tsp_bcr(parent_1, parent_2, list_of_functions):
    
    ################################################
    def ensure_all_numbers_present(lst, range_start, range_end):
        full_set        = set(list(range(range_start, range_end + 1)))
        lst_set         = set(lst)
        missing_numbers = full_set - lst_set
        lst.extend(missing_numbers)
        return lst
    
    def apply_functions_to_list(dist_list, list_of_functions):
        return [ (item, *[list_of_functions[k](item) for k in range(0, len(list_of_functions))]) for item in dist_list ]
    ################################################
   
    p_1    = copy.deepcopy(parent_1)
    p_2    = copy.deepcopy(parent_2)
    ct     = random.randint(0, len(p_2[0]) - 1)
    p_     = p_2[0][:ct + 1]  
    p_0    = [x for x in p_1[0] if x not in p_]
    p_2[0] = p_ + p_0
    M      = len(list_of_functions)
    jobs   = len(parent_1[0])
    cut    = random.sample(list(range(0, len(p_1[0]))), int(len(p_1)/2))
    cut    = [ p_1[0][i] for i in cut ]
    best   = []
    for j in range(0, len(cut)):
        A         = cut[j]
        p_2[0].remove(A)
        dist_list = []
        for i in range(0, len(p_2) + 1):
            new_list = p_2[0][:i] + [A] + p_2[0][i:]
            if (len(new_list) < jobs):
                new_list = ensure_all_numbers_present(new_list, 0, jobs-1)
            if (new_list not in dist_list):
                dist_list.append(new_list)
        merged = apply_functions_to_list(dist_list, list_of_functions)
        best.extend(merged)
    best = selection_leaders(1, M, best)
    ind  = []
    for m in range(0, len(best)):
        for n in range(0, len(best[m])):
            ind.append(best[m][n])
    return ind

# Function: Breeding
def breeding_ecmoa(leaders, population, list_of_functions):
    offspring = [[] for item in population]
    parent_1  = 0
    parent_2  = 1
    for i in range(0, len(leaders)):
        offspring[i] = leaders[i]
    for i in range (len(leaders), len(offspring)):
        i1, i2, i3, i4 = random.sample(range(0, len(population) - 1), 4)
        rand  = int.from_bytes(os.urandom(8), byteorder = 'big') / ((1 << 64) - 1)
        if (rand > 0.5):
            parent_1 = i1
        else:
            parent_1 = i2  
        rand  = int.from_bytes(os.urandom(8), byteorder = 'big') / ((1 << 64) - 1)
        if (rand > 0.5):
            parent_2 = i3
        else:
            parent_2 = i4
        parent_1 = copy.deepcopy(population[parent_1])  
        parent_2 = copy.deepcopy(population[parent_2])
        rand     = int.from_bytes(os.urandom(8), byteorder = 'big') / ((1 << 64) - 1)  
        if (rand > 0.5):
            offspring[i] = crossover_tsp_bcr(parent_1, parent_2, list_of_functions)
        else:
            offspring[i] = crossover_tsp_bcr(parent_2, parent_1, list_of_functions)
    return offspring

# Function: Mutation - Swap 
def mutation_ecmoa_swap(population, list_of_functions):
    p = copy.deepcopy(population)
    q = []
    for individual in p:
        k                 = random.sample(list(range(0, len(individual[0]))), 2)
        k1                = k[0]
        k2                = k[1]  
        A                 = individual[0][k1]
        B                 = individual[0][k2]
        individual[0][k1] = B
        individual[0][k2] = A
        for k in range (1, len(list_of_functions) + 1):
            individual[-k] = list_of_functions[-k](individual[0])
        q.append(individual)
    return q

############################################################################

# Function:  Pareto Front  
def pareto_front_points(pts, pf_min = True):
    
    ################################################
    def pareto_front(pts, pf_min):
        pf = np.zeros(pts.shape[0], dtype = np.bool_)
        for i in range(0, pts.shape[0]):
            cost = pts[i, :]
            if (pf_min == True):
                g_cost = np.logical_not(np.any(pts > cost, axis = 1))
                b_cost = np.any(pts < cost, axis = 1)
            else:
                g_cost = np.logical_not(np.any(pts < cost, axis = 1))
                b_cost = np.any(pts > cost, axis = 1)
            dominated = np.logical_and(g_cost, b_cost)
            if  (np.any(pf) == True):
                if (np.any(np.all(pts[pf] == cost, axis = 1)) == True):
                    continue
            if not (np.any(dominated[:i]) == True or np.any(dominated[i + 1 :]) == True):
                pf[i] = True
        return pf
    ################################################
    
    idx     = np.argsort(((pts - pts.mean(axis = 0))/(pts.std(axis = 0) + 1e-7)).sum(axis = 1))
    pts     = pts[idx]
    pf      = pareto_front(pts, pf_min)
    pf[idx] = pf.copy()
    return pf

############################################################################

# ECMOA Function
def elitist_combinatorial_multiobjective_optimization_algorithm(size = 15, jobs = 7, list_of_functions = [func_1, func_2], generations = 1500, k = 4, verbose = True):       
    count      = 0
    size       = max(5, size)
    M          = len(list_of_functions)
    k_size     = k*size
    population = initial_population_ecmoa(k_size, jobs, list_of_functions)  
    leaders    = selection_leaders(k_size, M, population)
    print(' Population Size: ', int(k_size))
    while (count <= generations):       
        if (verbose == True):
            print('Generation = ', count)
        offspring  = breeding_ecmoa(leaders, population, list_of_functions)
        mutants    = mutation_ecmoa_swap(population, list_of_functions)
        population = population + offspring + mutants
        population = ensure_unique_population(population)
        leaders    = selection_leaders(size, M, population)
        crowding   = crowding_distance_function(population, M)
        arg        = np.argsort(crowding , axis = 0)[::-1].tolist()
        try:
            arg = [i[0] for i in arg ]
        except:
            arg = [i for i in arg ]
        if (len(arg) > 0):
            population = [ population[idx] for idx in arg]
        population = population[:size]
        count      = count + 1              
    return leaders

############################################################################



############################################################################

# Created by: Prof. Valdecy Pereira, D.Sc.
# UFF - Universidade Federal Fluminense (Brazil)
# email:  valdecy.pereira@gmail.com
# GA Scheduler

# Citation: 
# PEREIRA, V. (2024). Project: GA Scheduler, GitHub repository: <https://github.com/Valdecy/GA_Scheduler>

############################################################################

# Required Libraries
import itertools
import math
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
plt.style.use('ggplot')
import numpy as np
import random

from ga_scheduler.util.ga    import genetic_algorithm
from ga_scheduler.util.ecmoa import elitist_combinatorial_multiobjective_optimization_algorithm, selection_leaders

############################################################################

# GA Scheduler Class
class load_ga_scheduler():
    def __init__(self, sequences = [], due_dates = [], setup_time_matrix = [], setup_waste_matrix = [], z_permutations = 100, job_weights = [], obj_makespan = True, obj_total_compl_time = True, obj_total_late_jobs = True, obj_max_w_tardiness = True, obj_total_waste = True, obj_setup = True, w_obj_makespan = 1, w_obj_total_compl_time = 1, w_obj_total_late_jobs = 1, w_obj_max_w_tardiness = 1, w_obj_total_waste = 1, w_obj_setup = 1, parallel_machines = False, brute_force = False, pareto_front = False, custom_sequence = []): 
      self.obj_1                          = obj_makespan
      self.obj_2                          = obj_max_w_tardiness
      self.obj_3                          = obj_total_waste
      self.obj_4                          = obj_setup
      self.obj_5                          = obj_total_compl_time
      self.obj_6                          = obj_total_late_jobs
      self.lst_func                       = []
      self.z_std                          = []
      self.z_permutations                 = z_permutations
      self.sequences                      = sequences                  # Job Shop Scheduling Input: Mandatory
      self.due_dates                      = due_dates                  # Job Shop Scheduling Input: Only Relevant if Due Date    is a measure
      self.setup_time_matrix              = setup_time_matrix          # Job Shop Scheduling Input: Only Relevant if Setup Time  is a measure
      self.setup_waste_matrix             = setup_waste_matrix         # Job Shop Scheduling Input: Only Relevant if Setup Waste is a measure
      self.parallel                       = parallel_machines
      self.brute_force                    = brute_force
      self.custom_sequence                = custom_sequence
      self.ecmo                           = pareto_front
      self.machine_sequences, self.matrix = self.sequence_inputs()
      self.num_jobs                       = len(self.sequences)
      self.num_machines                   = self.matrix.shape[1]
      self.job_weights                    = [1 for i in range(0, self.num_jobs)]
      self.objectives_weights             = [1, 1, 1, 1, 1, 1] 
      for i in range(0, min(len(job_weights), self.num_jobs)):
          self.job_weights[i] = job_weights[i] 
      if (self.obj_1 == True):
          self.objectives_weights[0] = w_obj_makespan
      else:
          self.objectives_weights[0] = 0
      if (self.obj_2 == True):
          self.objectives_weights[1] = w_obj_max_w_tardiness
      else:
          self.objectives_weights[1] = 0
      if (self.obj_3 == True):
          self.objectives_weights[2] = w_obj_total_waste
      else:
          self.objectives_weights[2] = 0
      if (self.obj_4 == True):
          self.objectives_weights[3] = w_obj_setup
      else:
          self.objectives_weights[3] = 0
      if (self.obj_5 == True):
          self.objectives_weights[4] = w_obj_total_compl_time
      else:
          self.objectives_weights[4] = 0
      if (self.obj_6 == True):
          self.objectives_weights[5] = w_obj_total_late_jobs
      else:
          self.objectives_weights[5] = 0
      if (self.objectives_weights.count(0) == len(self.objectives_weights) -1):
        self.z_mean, self.z_std = [0]*len(self.objectives_weights), [1]*len(self.objectives_weights) 
      else:
        self.z_mean, self.z_std = self.obj_z_search()
      self.objective_functions()
    ###############################################################################   
    
    # Create Instances
    def generate_sequences(self, num_jobs, num_machines, use_all_machines = False, parallel_machines = False, parallel_same_time = False, flow_shop = False, bounds = [1, 10], seed = None):
        
        ################################################
        def process_lists(lists, pm = False, tm = False):
            processed_lists = []
            for lst in lists:
                if (pm):
                    lst = sorted(lst, key = lambda x: x[0])
                if (tm):
                    first_tuple_second_element = lst[0][1]
                    lst = [(t[0], first_tuple_second_element) for t in lst]
                processed_lists.append(lst)
            return processed_lists
        ################################################
        
        if (seed is not None):
            random.seed(seed)
        sequences = []
        if (flow_shop == True):
            base_sequence = random.sample(range(0, num_machines), num_machines)
        for job in range(0, num_jobs):
            job_sequence = []
            if (flow_shop == True):
                for machine in base_sequence:
                    process_time = random.randint(bounds[0], bounds[1])
                    job_sequence.append((machine, process_time))
            else:
                if (use_all_machines):
                    machine_sequence = random.sample(range(0, num_machines), num_machines)
                else:
                    machine_sequence = random.sample(range(0, num_machines), random.randint(1, num_machines))
                for machine in machine_sequence:
                    process_time = random.randint(bounds[0], bounds[1])
                    job_sequence.append((machine, process_time))
            sequences.append(job_sequence)
        if (parallel_machines == True or parallel_same_time == True):
            sequences = process_lists(lists = sequences, pm = parallel_machines, tm = parallel_same_time)
        return sequences
    
    # Schedule
    def schedule_jobs(self, permutation):
        total_length      = np.sum(self.matrix)
        schedule          = [['' for _ in range(0, total_length)] for _ in range(0, self.num_machines)]
        machine_end_times = [0] * self.num_machines
        job_end_times     = [0] * self.num_jobs
        if (self.parallel == False):
            for job_id in permutation:
                operations = self.machine_sequences[job_id]
                for op_index, machine in enumerate(operations):
                    time_required = self.matrix[job_id, machine]
                    start_time    = job_end_times[job_id]
                    while any(schedule[machine][start_time:start_time + time_required]):
                        start_time = start_time + 1
                    end_time = start_time + time_required
                    for t in range(start_time, end_time):
                        schedule[machine][t] = f"j{job_id}"
                    job_end_times[job_id]      = end_time
                    machine_end_times[machine] = end_time
        else:        
            for job_id in permutation:
                earliest_start_time = float('inf')
                best_machine        = None
                for machine, time_required in self.sequences[job_id]:
                    start_time = machine_end_times[machine]
                    while any(schedule[machine][start_time:start_time + time_required]):
                        start_time = start_time + 1
                    if (start_time < earliest_start_time):
                        earliest_start_time = start_time
                        best_machine        = machine
                time_required = self.matrix[job_id, best_machine]
                end_time      = earliest_start_time + time_required
                for t in range(earliest_start_time, end_time):
                    schedule[best_machine][t]   = f"j{job_id}"
                machine_end_times[best_machine] = end_time
                job_end_times[job_id]           = end_time
        max_time = max(len(row) for row in schedule)
        for col in range(max_time - 1, -1, -1):
            if (all(row[col] == '' for row in schedule)):
                for row in schedule:
                    row.pop()
            else:
                break
        return np.array(schedule)
    
    ###############################################################################
    
    # Plot
    def create_gantt_chart(self, schedule_matrix, size_x = 12, size_y = 8):
        non_empty_rows = [i for i in range(schedule_matrix.shape[0]) if not np.all(schedule_matrix[i] == '')]
        schedule       = schedule_matrix[non_empty_rows, :]
        num_machines   = schedule.shape[0]
        total_length   = schedule.shape[1]
        fig, ax        = plt.subplots(figsize = (size_x, size_y))
        job_color_dict = {}
        idle_patches   = []
        job_colors     = [
                            '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', 
                            '#7f7f7f', '#bcbd22', '#17becf', '#bf77f6', '#ff9408', '#d1ffbd', '#c85a53', 
                            '#3a18b1', '#ff796c', '#04d8b2', '#ffb07c', '#aaa662', '#0485d1', '#fffe7a', 
                            '#b0dd16', '#6f7be3', '#12e193', '#82cafc', '#ac9362', '#f8481c', '#c292a1', 
                            '#c0fa8b', '#ca7b80', '#f4d054', '#fbdd7e', '#ffff7e', '#cd7584', '#f9bc08', 
                            '#c7c10c'
                         ]
        for machine in range(0, num_machines):
            for time in range(0, total_length):
                job = schedule[machine][time]
                if (job):
                    if (job not in job_color_dict):
                        job_index           = int(job[1:])  
                        job_color_dict[job] = job_colors[job_index % len(job_colors)]
                    color = job_color_dict[job]
                    ax.add_patch(mpatches.Rectangle((time, num_machines - machine - 1), 1, 1, edgecolor = 'black', facecolor = color))
                    ax.text(time + 0.5, num_machines - machine - 0.5, job, ha = 'center', va = 'center', color = 'black', fontsize = 8, weight = 'bold')
                else:
                    left_hatch = mpatches.Rectangle((time, num_machines - machine - 1), 1, 1, edgecolor = 'black', facecolor = 'none', hatch = '//')
                    right_hatch = mpatches.Rectangle((time, num_machines - machine - 1), 1, 1, edgecolor = 'black', facecolor = 'none', hatch = '\\')
                    ax.add_patch(left_hatch)
                    ax.add_patch(right_hatch)
                    idle_patches.append((time, machine, left_hatch, right_hatch))
        for machine in range(0, num_machines):
            for time in range(total_length - 1, -1, -1):
                if (schedule[machine][time] == ''):
                    for patch in idle_patches:
                        if (patch[0] == time and patch[1] == machine):
                            patch[2].remove()
                            patch[3].remove()
                else:
                    break
        ax.set_xlim(0, total_length)
        ax.set_ylim(0, num_machines)
        ax.set_xticks(np.arange(0, total_length + 1, 1))
        ax.set_xticklabels(np.arange(0, total_length + 1, 1))
        ax.set_yticks(np.arange(0, num_machines, 1))
        ax.set_yticklabels([])
        for i in range(0, num_machines):
            ax.text(-0.5, num_machines - i - 0.5, f'Machine {i}', va = 'center', ha = 'right', fontsize = 10)
        ax.set_xlabel('Time')
        ax.set_title('Gantt Chart')
        ax.grid(True, linestyle = '--', alpha = 0.7)
        plt.show()
        
    ###############################################################################
    
    # Objectives
    def calculate_makespan(self, schedule_matrix):
        total_length = schedule_matrix.shape[1]
        makespan     = 0
        for machine in range(0, self.num_machines):
            for time in range(total_length - 1, -1, -1):
                if (schedule_matrix[machine][time] != ''):
                    makespan = max(makespan, time + 1)
                    break
        return makespan

    def calculate_makespan_p(self, permutation):
        schedule_matrix = self.schedule_jobs(permutation)
        total_length    = schedule_matrix.shape[1]
        makespan        = 0
        for machine in range(0, self.num_machines):
            for time in range(total_length - 1, -1, -1):
                if (schedule_matrix[machine][time] != ''):
                    makespan = max(makespan, time + 1)
                    break
        return makespan
    
    def calculate_max_weighted_tardiness(self, schedule_matrix):
        completion_times = {f'j{job}': 0 for job in range(0, self.num_jobs)}
        total_length     = schedule_matrix.shape[1]
        for machine in range(0, self.num_machines):
            for time in range(0, total_length):
                job = schedule_matrix[machine][time]
                if (job):
                    completion_times[job] = time + 1
        max_weighted_tardiness = 0
        for job in range(0, self.num_jobs):
            tardiness              = max(0, completion_times[f'j{job}'] - self.due_dates[job])
            weighted_tardiness     = self.job_weights[job] * tardiness
            max_weighted_tardiness = max(max_weighted_tardiness, weighted_tardiness)
        return max_weighted_tardiness

    def calculate_max_weighted_tardiness_p(self, permutation):
        schedule_matrix  = self.schedule_jobs(permutation)
        completion_times = {f'j{job}': 0 for job in range(0, self.num_jobs)}
        total_length     = schedule_matrix.shape[1]
        for machine in range(0, self.num_machines):
            for time in range(0, total_length):
                job = schedule_matrix[machine][time]
                if (job):
                    completion_times[job] = time + 1
        max_weighted_tardiness = 0
        for job in range(0, self.num_jobs):
            tardiness              = max(0, completion_times[f'j{job}'] - self.due_dates[job])
            weighted_tardiness     = self.job_weights[job] * tardiness
            max_weighted_tardiness = max(max_weighted_tardiness, weighted_tardiness)
        return max_weighted_tardiness

    def calculate_total_completion_time(self, schedule_matrix):
        completion_times = {f'j{job}': 0 for job in range(0, self.num_jobs)}
        total_length     = schedule_matrix.shape[1]
        for machine in range(0, self.num_machines):
            for time in range(0, total_length):
                job = schedule_matrix[machine][time]
                if (job):  
                    completion_times[job] = time + 1  
        total_completion_time = sum(completion_times[f'j{job}'] for job in range(0, self.num_jobs))
        return total_completion_time  

    def calculate_total_completion_time_p(self, permutation):
        schedule_matrix  = self.schedule_jobs(permutation)
        completion_times = {f'j{job}': 0 for job in range(0, self.num_jobs)}
        total_length     = schedule_matrix.shape[1]
        for machine in range(0, self.num_machines):
            for time in range(0, total_length):
                job = schedule_matrix[machine][time]
                if (job):  
                    completion_times[job] = time + 1  
        total_completion_time = sum(completion_times[f'j{job}'] for job in range(0, self.num_jobs))
        return total_completion_time  

    def calculate_total_late_jobs(self, schedule_matrix):
        completion_times = {f'j{job}': 0 for job in range(0, self.num_jobs)}
        total_length     = schedule_matrix.shape[1]
        for machine in range(0, self.num_machines):
            for time in range(0, total_length):
                job = schedule_matrix[machine][time]
                if (job): 
                    completion_times[job] = time + 1
        total_late_jobs = 0
        for job in range(0, self.num_jobs):
            if (completion_times[f'j{job}'] > self.due_dates[job]):  
                total_late_jobs = total_late_jobs + 1  
        return total_late_jobs  
    
    def calculate_total_late_jobs_p(self, permutation):
        schedule_matrix  = self.schedule_jobs(permutation)
        completion_times = {f'j{job}': 0 for job in range(0, self.num_jobs)}
        total_length     = schedule_matrix.shape[1]
        for machine in range(0, self.num_machines):
            for time in range(0, total_length):
                job = schedule_matrix[machine][time]
                if (job): 
                    completion_times[job] = time + 1
        total_late_jobs = 0
        for job in range(0, self.num_jobs):
            if (completion_times[f'j{job}'] > self.due_dates[job]):  
                total_late_jobs = total_late_jobs + 1  
        return total_late_jobs
    
    def calculate_total_waste(self, permutation):
        total_waste = 0
        for i in range(0, len(permutation) - 1):
            total_waste = total_waste + self.setup_waste_matrix[permutation[i]][permutation[i+1]]
        return total_waste

    def calculate_idle_times(self, schedule_matrix):
        total_idle_time = 0
        for row in schedule_matrix:
            row_idle_time  = 0
            last_job_index = None
            for i in range(len(row) - 1, -1, -1):
                if (row[i] != ''):
                    last_job_index = i
                    break
            if (last_job_index is not None):
                for i in range(0, last_job_index):
                    if (row[i] == ''):
                        row_idle_time =  row_idle_time  + 1
            total_idle_time = total_idle_time + row_idle_time
        return total_idle_time
    
    def calculate_idle_times_p(self, permutation):
        schedule_matrix = self.schedule_jobs(permutation)
        total_idle_time = 0
        for row in schedule_matrix:
            row_idle_time  = 0
            last_job_index = None
            for i in range(len(row) - 1, -1, -1):
                if (row[i] != ''):
                    last_job_index = i
                    break
            if (last_job_index is not None):
                for i in range(0, last_job_index):
                    if (row[i] == ''):
                        row_idle_time =  row_idle_time  + 1
            total_idle_time = total_idle_time + row_idle_time
        return total_idle_time
    
    def calculate_total_setup_time(self, permutation):
        total_setup_time = 0
        for i in range(0, len(permutation) - 1):
            total_setup_time = total_setup_time + self.setup_time_matrix[permutation[i]][permutation[i+1]]
        return total_setup_time
    
    ###############################################################################
    
    # Inputs
    def sequence_inputs(self):
        num_jobs     = len(self.sequences)
        num_machines = max(max(machine for machine, _ in job) for job in self.sequences) + 1
        matrix       = np.zeros((num_jobs, num_machines), dtype = int)
        for job_id, job in enumerate(self.sequences):
            for machine, time in job:
                matrix[job_id, machine] = time
        machine_sequences = []
        for job_id, job in enumerate(self.sequences):
            machine_sequence = [machine for machine, _ in job]
            machine_sequences.append(machine_sequence)
        return machine_sequences, matrix
    
    ###############################################################################
    
    # Find Solution
    def brute_force_search(self):
        job_ids                 = list(range(0, len(self.sequences)))
        best_sequence           = None
        minimal_objective_value = float('inf')
        count                   = 1
        total                   = math.factorial(len(job_ids))
        print('')
        print('Verifying', total, 'solutions')
        print('')
        for permutation in itertools.permutations(job_ids):
            objective_value = self.target_function(permutation)
            count           = count + 1
            searched_perc   = round(count / total * 100, 4)
            if (objective_value < minimal_objective_value):
                minimal_objective_value = objective_value
                best_sequence           = permutation
                print(f"Searched Space: {searched_perc:.2f}%; Best Sequence: {best_sequence}; Obj. Function: {minimal_objective_value:.4f}")
        print('')
        print('Brute Force Search Complete!')
        print('')
        print(f"Best Sequence: {best_sequence}; Obj. Function: {minimal_objective_value:.4f}")
        self.best_sequence = best_sequence
        return best_sequence, minimal_objective_value
    
    def brute_force_search_p(self):
        merged  = [] 
        job_ids = list(range(0, len(self.sequences)))
        count   = 1
        total   = math.factorial(len(job_ids))
        print('')
        print('Verifying', total, 'solutions')
        print('')
        for permutation in itertools.permutations(job_ids):
            m = []
            p = [item for item in permutation]
            m.append(p)
            for k in range(0, len(self.lst_func)):
                m.append(self.lst_func[k](p))
            merged.append(m)
            count         = count + 1
            searched_perc = round(count / total * 100, 4)
            if (count % (int(total/10)) == 0 or count == total):
                print(f"Searched Space: {searched_perc:.2f}%")
        print('Brute Force Search Complete!')
        leaders = selection_leaders(total, len(self.lst_func), merged)
        return leaders
    
    def obj_z_search(self):
        
        ################################################
        def unique_permutations(job_ids, num_permutations):
            seen = set()
            while (len(seen) < num_permutations):
                perm = tuple(random.sample(job_ids, len(job_ids)))
                seen.add(perm)
            return list(seen)
        ################################################
        
        job_ids              = list(range(0, len(self.sequences)))
        num_permutations     = min(self.z_permutations, math.factorial(len(job_ids)))
        sampled_permutations = unique_permutations(job_ids, num_permutations)
        obj_1_lst            = []
        obj_2_lst            = []
        obj_3_lst            = []
        obj_4_lst            = []
        obj_5_lst            = []
        obj_6_lst            = []
        z_mean               = []
        z_std                = []
        for permutation in sampled_permutations:
            schedule_matrix = self.schedule_jobs(permutation)
            if (self.obj_1 == True):
                makespan = self.calculate_makespan(schedule_matrix)
                obj_1_lst.append(makespan)
            if (self.obj_2 == True):
                max_weighted_tardiness = self.calculate_max_weighted_tardiness(schedule_matrix)
                obj_2_lst.append(max_weighted_tardiness)
            if (self.obj_3 == True):
                total_waste = self.calculate_total_waste(permutation)
                obj_3_lst.append(total_waste)
            if (self.obj_4 == True):
                total_setup_time = self.calculate_total_setup_time(permutation)
                obj_4_lst.append(total_setup_time)
            if (self.obj_5 == True):
                total_compl_time = self.calculate_total_completion_time(schedule_matrix)
                obj_5_lst.append(total_compl_time)
            if (self.obj_6 == True):
                total_late_jobs = self.calculate_total_late_jobs(schedule_matrix)
                obj_6_lst.append(total_late_jobs)
        if (obj_1_lst):
            z_mean.append(np.mean(obj_1_lst))
            #z_std.append(np.std(obj_1_lst, ddof = 1) / np.sqrt(len(obj_1_lst)))
            z_std.append(np.std(obj_1_lst, ddof = 1))
        else:
            z_mean.append(None)
            z_std.append(None)
        if (obj_2_lst):
            z_mean.append(np.mean(obj_2_lst))
            #z_std.append(np.std(obj_2_lst, ddof = 1) / np.sqrt(len(obj_2_lst)))
            z_std.append(np.std(obj_2_lst, ddof = 1))
        else:
            z_mean.append(None)
            z_std.append(None)
        if (obj_3_lst):
            z_mean.append(np.mean(obj_3_lst))
            #z_std.append(np.std(obj_3_lst, ddof = 1) / np.sqrt(len(obj_3_lst)))
            z_std.append(np.std(obj_3_lst, ddof = 1))
        else:
            z_mean.append(None)
            z_std.append(None)
        if (obj_4_lst):
            z_mean.append(np.mean(obj_4_lst))
            #z_std.append(np.std(obj_4_lst, ddof = 1) / np.sqrt(len(obj_4_lst)))
            z_std.append(np.std(obj_4_lst, ddof = 1))
        else:
            z_mean.append(None)
            z_std.append(None)
        if (obj_5_lst):
            z_mean.append(np.mean(obj_5_lst))
            #z_std.append(np.std(obj_5_lst, ddof = 1) / np.sqrt(len(obj_5_lst)))
            z_std.append(np.std(obj_5_lst, ddof = 1))
        else:
            z_mean.append(None)
            z_std.append(None)
        if (obj_6_lst):
            z_mean.append(np.mean(obj_6_lst))
            #z_std.append(np.std(obj_6_lst, ddof = 1) / np.sqrt(len(obj_6_lst)))
            z_std.append(np.std(obj_6_lst, ddof = 1))
        else:
            z_mean.append(None)
            z_std.append(None)
        return z_mean, z_std
    
    def target_function(self, permutation): 
        schedule_matrix = self.schedule_jobs(permutation)
        objective_value = 0.0 / 1.0
        if (self.objectives_weights[0] != 0):
            makespan                 = self.calculate_makespan(schedule_matrix)
            z_makespan               = ( (makespan - self.z_mean[0]) / (self.z_std[0] + 1e-14)) + 9
            objective_value          = objective_value + self.objectives_weights[0] * z_makespan
        if (self.objectives_weights[1] != 0):
            max_weighted_tardiness   = self.calculate_max_weighted_tardiness(schedule_matrix)
            z_max_weighted_tardiness = ( (max_weighted_tardiness - self.z_mean[1]) /  (self.z_std[1] + 1e-14) ) + 9
            objective_value          = objective_value + self.objectives_weights[1] * z_max_weighted_tardiness
        if (self.objectives_weights[2] != 0):
            total_waste              = self.calculate_total_waste(permutation)
            z_total_waste            = ( (total_waste - self.z_mean[2]) /  (self.z_std[2] + 1e-14) ) + 9
            objective_value          = objective_value + self.objectives_weights[2] * z_total_waste 
        if (self.objectives_weights[3] != 0):
            total_setup_time         = self.calculate_total_setup_time(permutation)
            z_total_setup_time       = ( (total_setup_time - self.z_mean[3]) /  (self.z_std[3] + 1e-14) ) + 9
            objective_value          = objective_value + self.objectives_weights[3] * z_total_setup_time 
        if (self.objectives_weights[4] != 0):
            total_compl_time         = self.calculate_total_completion_time(schedule_matrix)
            z_total_compl_time       = ( (total_compl_time - self.z_mean[4]) /  (self.z_std[4] + 1e-14) ) + 9
            objective_value          = objective_value + self.objectives_weights[4] * z_total_compl_time 
        if (self.objectives_weights[5] != 0):
            total_late_jobs          = self.calculate_total_late_jobs(schedule_matrix)
            z_total_late_jobs        = ( (total_late_jobs - self.z_mean[5]) /  (self.z_std[5] + 1e-14) ) + 9
            objective_value          = objective_value + self.objectives_weights[5] * z_total_late_jobs 
        return objective_value
    
    def objective_functions(self):
        if (self.obj_1 == True):
            self.lst_func.append(self.calculate_makespan_p)
        if (self.obj_2 == True):
            self.lst_func.append(self.calculate_max_weighted_tardiness_p)
        if (self.obj_3 == True): 
            self.lst_func.append(self.calculate_total_waste) 
        if (self.obj_4 == True):
            self.lst_func.append(self.calculate_total_setup_time)
        if (self.obj_5 == True):
            self.lst_func.append(self.calculate_total_completion_time_p)
        if (self.obj_6 == True):
            self.lst_func.append(self.calculate_total_late_jobs_p)
        return 
    
    ###############################################################################
    
    # GA Scheduler
    def run_ga_scheduler(self, population_size = 5, elite = 1, mutation_rate = 0.10, generations = 100, k = 4):
        if (self.brute_force == False and len(self.custom_sequence) == 0 and self.ecmo == False):
            job_sequence, obj_fun = genetic_algorithm(self.num_jobs, population_size, elite, mutation_rate, generations, self.target_function, True)
            schedule_matrix       = self.schedule_jobs(job_sequence) 
            return job_sequence, schedule_matrix, obj_fun
        elif (self.brute_force == False and len(self.custom_sequence) == 0 and self.ecmo == True):
            leaders = elitist_combinatorial_multiobjective_optimization_algorithm(population_size, self.num_jobs, self.lst_func, generations, k, True)
            return leaders
        if (self.brute_force == True and len(self.custom_sequence) == 0 and self.ecmo == False):
            job_sequence, obj_fun = self.brute_force_search()
            schedule_matrix       = self.schedule_jobs(job_sequence)
            return job_sequence, schedule_matrix, obj_fun
        elif (self.brute_force == True and len(self.custom_sequence) == 0 and self.ecmo == True):
            leaders = self.brute_force_search_p()
            return leaders
        if (len(self.custom_sequence) != 0):
            job_sequence    = self.custom_sequence
            schedule_matrix = self.schedule_jobs(job_sequence) 
            obj_fun         = self.target_function(self.custom_sequence)
            return job_sequence, schedule_matrix, obj_fun
    
############################################################################



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()

setup(
    name='ga_scheduler',
    version='2.7.8',
    license='GNU',
    author='Valdecy Pereira',
    author_email='valdecy.pereira@gmail.com',
    url='https://github.com/Valdecy/ga_scheduler',
    packages=find_packages(),
    install_requires=[
        'matplotlib',
        'numpy'
    ],
    description='A Comprehensive Library for Solving Machine Scheduling Problems Using Genetic Algorithms',
    long_description=long_description,
    long_description_content_type='text/markdown',
)